from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional

from openai import OpenAI

if TYPE_CHECKING:
    import pydantic


@dataclass
class Verdict:
    """
    A dataclass that represents a judge's verdict.
    """
    score: bool | int
    reasoning: str


class BaseJudge:
    """
    Base class for all judges. All judges must implement a judge method which
    produces a `Verdict` object that contains a score and reasoning.

    The score is a boolean value, True if judge's criteria is met, False
    if not.

    The reasoning is a string that explains why the score is True or False.
    """
    citation: str = None

    def __init__(
        self,
        model: str,
    ):
        """
        Initialize the judge with a specific model.
        """
        self.model = model
        self._client = self._configure_client()

    def __init_subclass__(cls, **kwargs):
        for required in ("citation",):
            if not getattr(cls, required):
                raise TypeError(f"can't instantiate abstract class {cls.__name__} without {required} attribute defined")
        return super().__init_subclass__(**kwargs)

    def _configure_client(self):
        if self.model == "gpt-4o-mini":
            client = OpenAI()
        elif self.model == "gpt-4o":
            client = OpenAI()
        else:
            raise ValueError(f"Invalid model: {self.model}")

        return client
    
    def _get_completion(self, user_prompt: str, system_prompt: Optional[str] = None):
        messages = []
        if system_prompt is not None:
            messages.append({"role": "system", "content": system_prompt})
        
        messages.append({"role": "user", "content": user_prompt})

        completion = self._client.chat.completions.create(
            model=self.model,
            messages=messages,
            response_format={"type": "json_object"}
        )
        return completion

    def judge(
        self,
        input: str,
        output: str = None,
        expected: str = None,
        format: Optional["pydantic.BaseModel"] = None,
    ) -> Verdict:
        """
        Judge the input and return a verdict.
        """
        raise NotImplementedError("All judges must implement a judge method")



from functools import partial
from textwrap import dedent
from typing import Optional

from judges.base import BaseJudge, Verdict

class CorrectnessJudge(BaseJudge):
    """
    Judge that determines if the output is correct.
    """
    citation = dedent(r"""
        @article{FILLMEIN,
            title={FILLMEIN},
            author={FILLMEIN},
            journal={FILLMEIN},
            year={FILLMEIN},
            volume={FILLMEIN},
            url={FILLMEIN}
        }
    """)
    def judge(
        self,
        input: str,
        output: str = None,
        expected: str = None,
        format: Optional["pydantic.BaseModel"] = None,
    ) -> Verdict:
        """
        Judge the input and return a verdict.
        """
        system_prompt = None
        user_prompt = dedent(f"""
        You are given a QUESTION, an ANSWER and a REFERENCE. 
        You must determine whether the given ANSWER correctly answers the QUESTION based on the REFERENCE. 
                             
        Here is the data:
                             
        [BEGIN DATA]
        ************
        [Question]: {input}
        ************
        [Reference]: {expected}
        ************
        [Answer]: {output}
        [END DATA]

        Your response must be a single word, either "correct" or "incorrect",
        and should not contain any text or characters aside from that word.

        "correct" means that the question is correctly and fully answered by the answer.
        "incorrect" means that the question is not correctly or only partially answered by the
        answer.
        """)


correctness = partial(CorrectnessJudge)


import os
from unittest.mock import patch

import openai

from judges.base import BaseJudge, Verdict
from judges.classifiers.correctness import CorrectnessJudge
from judges.graders.quality import QueryQualityJudge
from judges.graders.relevancy import RelevancyJudge

os.environ['OPENAI_API_KEY'] = 'test-key'

def mock_verdict(verdict: Verdict):
    def decorator(func):
        def wrapper(*args, **kwargs):
            openai.api_type = 'openai'
            with patch("openai.chat.completions.create") as mock:
                mock.return_value = verdict
                 # Pass the mock to the test
                return func(mock, *args, **kwargs)
        return wrapper
    return decorator


def test_base_judge_no_citation():


    try:
        class TestJudge(BaseJudge):
            pass
    except TypeError as e:
        assert str(e) == "can't instantiate abstract class TestJudge without citation attribute defined"

def test_base_judge():
    class TestJudge(BaseJudge):
        citation = "Test"

    judge = TestJudge("gpt-4o-mini")
    assert judge.model == "gpt-4o-mini"
    assert judge._client is not None


# @mock_verdict(
#     Verdict(
#         score="average",
#         reasoning="The query is moderately clear and specific. It may require some additional information for a complete understanding.",
#     )
# )
# def test_quality_judge(mockclient):
#     judge = QueryQualityJudge(model="gpt-4o-mini")
#     verdict = judge.judge(input="What is the capital of France?")

#     assert verdict.score == "average"
#     assert (
#         verdict.reasoning
#         == "The query is moderately clear and specific. It may require some additional information for a complete understanding."
#     )
#     mockclient.assert_called_once()


# @mock_verdict(
#     Verdict(
#         score="correct",
#         reasoning="The answer correctly and fully answers the question.",
#     )
# )
# def test_correctness_judge(mockclient):
#     judge = CorrectnessJudge(model="gpt-4o-mini")
#     verdict = judge.judge(
#         input="What is the capital of France?",
#         output="Paris",
#         expected="Paris is the capital of France.",
#     )

#     assert verdict.score == "correct"
#     assert verdict.reasoning == "The answer correctly and fully answers the question."

#     mockclient.assert_called_once()


# @mock_verdict(
#     Verdict(
#         score=2,
#         reasoning="The question is highly relevant to the document.",
#     )
# )
# def test_relevancy_judge(mockclient):
#     judge = RelevancyJudge(model="gpt-4o-mini")
#     verdict = judge.judge(
#         input="What is the capital of France?",
#         output="Paris is the capital of France.",
#     )

#     assert verdict.score == 2
#     assert verdict.reasoning == "The question is highly relevant to the document."
#     mockclient.assert_called_once()


from judges.classifiers.correctness import correctness
from judges.graders.quality import query_quality
from judges.graders.relevancy import relevancy

__all__ = ["correctness", "query_quality", "relevancy"]


from functools import partial
from textwrap import dedent
from typing import Optional

from judges.base import BaseJudge, Verdict


class RelevancyJudge(BaseJudge):
    citation = dedent(r"""
        @misc{vespa2024llmjudge,
        author       = {Jo Kristian Bergum},
        title        = {Improving Retrieval with LLM as a Judge},
        howpublished = {\url{https://blog.vespa.ai/improving-retrieval-with-llm-as-a-judge/}},
        note         = {Accessed: 2024-09-17},
        year         = {2024}
        }
    """)

    def judge(
        self,
        input: str,
        output: str = None,
        expected: str = None,
        format: Optional["pydantic.BaseModel"] = None,
    ) -> Verdict:
        """
        Judge the input and return a verdict.
        """
        system_prompt = None
        user_prompt = dedent(f"""
        You are tasked with creating a system that can detect the relevance of a DOCUMENT to a QUESTION.

        Given a DOCUMENT and QUESTION you must provide a SCORE on an integer scale of 0 to 2 with the following meanings:

        0 = Represents that the QUESTION is irrelevant to the DOCUMENT
        1 = Represents that the QUESTION is somewhat relevant to the DOCUMENT
        2 = Represents that the QUESTION is highly relevant to the DOCUMENT

        Show your reasoning.

        ### Instruction
        Assign a relevancy score of 1 if the QUESTION is somewhat related to the DOCUMENT and CONTEXT, a score of 2 if QUESTION is highly relevant. 
        If none of the above satisfies give it a score of 0.

        ### Input
        QUESTION: {input}
        DOCUMENT: {output}

        Your output should be in JSON FORMAT with the keys "REASONING" and "SCORE":
        {{"REASONING": <your reasoning as bullet points>, "SCORE": <your final score>}}
        """
        )
        completion = self._get_completion(
            user_prompt=user_prompt,
            system_prompt=system_prompt,
        )
        reasoning = completion.choices[0].message.content["reasoning"]
        score = completion.choices[0].message.content["score"]
        return Verdict(score=score, reasoning=reasoning)

relevancy = partial(RelevancyJudge)